/*
 * process, Distribution repositories and basic setup for Emmanuel Bigeon projects
 * Copyright (C) 2014-2018 E. Bigeon
 * mailto:emmanuel@bigeon.fr
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/**
 * gclc:fr.bigeon.gclc.proc.TaskPool.java
 * Created on: May 10, 2017
 */
package net.bigeon.gclc.process;

/*-
 * #%L
 * process
 * %%
 * Copyright (C) 2014 - 2018 Bigeon
 * %%
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 * #L%
 */
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

/** A process pool.
 *
 * @author Emmanuel Bigeon */
public final class TaskPool {
    /** The running processes. */
    private final Map<String, Task> running = new HashMap<>();
    /** The count for process id. */
    private int                     count   = 0;
    /** The lock for pid attribution synchronization. */
    private final Object            lock    = new Object();
    private final boolean           autoClear;

    /** Default constructor. */
    public TaskPool() {
        this(true);
    }

    /** Default constructor. */
    public TaskPool(final boolean autoClear) {
        this.autoClear = autoClear;
    }

    /** Add a process in the pool.
     *
     * @param cmd the process
     * @return the pid */
    public String add(final Task cmd) {
        if (cmd == null) {
            throw new IllegalArgumentException("Task cannot be null"); //$NON-NLS-1$
        }
        final String pid;
        synchronized (lock) {
            pid = getPID();
            running.put(pid, cmd);
        }
        if (autoClear) {
            cmd.addInterruptionListener(new InterruptionListener() {
                @Override
                public void interrupted() {
                    remove(pid);
                    cmd.rmInterruptionListener(this);
                }
            });
        }
        return pid;
    }

    /** Get a process by it associated identifier.
     *
     * @param pid the task id
     * @return the task, if any, associated to this id */
    public Task get(final String pid) {
        synchronized (lock) {
            return running.get(pid);
        }
    }

    /** Get the next process id.
     *
     * @return the process id */
    private String getPID() {
        synchronized (lock) {
            String pid;
            do {
                pid = Integer.toString(count++);
            } while (running.containsKey(pid));
            return pid;
        }
    }

    /** Get the running processes' identifiers.
     *
     * @return the pids */
    public Collection<String> getPIDs() {
        return new HashSet<>(running.keySet());
    }

    /** @return if the clearing of ended task is automatic */
    public boolean isAutoClearing() {
        return autoClear;
    }

    /** Remove a task from the pool
     *
     * @param pid the task id */
    public void remove(final String pid) {
        synchronized (lock) {
            running.remove(pid);
            count = Math.min(count, Integer.parseInt(pid));
        }
    }

    /** Request all task to stop running.
     * <p>
     * This call does not guaranty end of execution, it is up to the task
     * implementation to actually take into account the termination request. */
    public void shutdown() {
        synchronized (lock) {
            for (final Task task : running.values()) {
                task.setRunning(false);
            }
        }
    }
}
